from dataclasses import dataclass


@dataclass
class Song:
    title: str
    album_name: str
    artist: str
    url: str
    content_type: str
    album_art: str

    def __str__(self) -> str:
        return f"{self.title} / {self.album_name} by {self.artist}"



from termcolor import cprint

_DEBUG = False


def enable_debug_mode():
    global _DEBUG  # noqa: PLW0603
    _DEBUG = True


def debug_mode_enabled() -> bool:
    return _DEBUG


def debug(context: str, msg: str):
    if _DEBUG:
        cprint(f"  [{context}] {msg}", "green")


def message(msg: str):
    print(msg)


def error(msg: str):
    cprint(msg, "red", attrs=["bold"])



from abc import abstractmethod


class NoSongsToPlayException(Exception):
    pass


class Backend:
    @abstractmethod
    def force_play(self):
        """Force playing the first song in the queue"""

    @abstractmethod
    def playpause(self):
        """Play or pause the music"""

    @abstractmethod
    def volume_set(self, value: float):
        """Set volume, between 0 and 1.0"""

    @abstractmethod
    def volume_delta(self, value: float):
        """add value to the volume, between 0 and 1.0"""

    @abstractmethod
    def stop(self):
        """Stop the music, regardless of its current status"""



import difflib
import random
import string
from hashlib import md5
from typing import Any, Dict, List, Tuple
from urllib.parse import urlencode

import requests

from castme.song import Song


class AlbumNotFoundException(Exception):
    def __init__(self, keyword: str):
        self.keyword = keyword

    def __str__(self):
        return f"Album not found with keyword: {self.keyword}"


SUBSONIC_SUPPORTED_VERSION = "1.16.1"


class SubSonic:
    """API client implementation based on the official documentation:
    https://www.subsonic.org/pages/api.jsp
    """

    def __init__(
        self, app_id: str, user: str, password: str, server_prefix: str
    ) -> None:
        self.app_id = app_id
        self.user = user
        self.password = password
        self.server_prefix = server_prefix

    def make_sonic_url(
        self, verb: str, **kwargs: str | int
    ) -> Tuple[str, Dict[str, Any]]:
        salt = "".join(random.choices(string.ascii_letters + string.digits, k=10))
        token = md5((self.password + salt).encode()).hexdigest()
        parameters = kwargs | {
            "u": self.user,
            "t": token,
            "v": SUBSONIC_SUPPORTED_VERSION,
            "c": self.app_id,
            "f": "json",
            "s": salt,
        }

        return f"{self.server_prefix}/rest/{verb}", parameters

    def call_sonic(self, verb: str, **kwargs: str | int):
        url, parameters = self.make_sonic_url(verb, **kwargs)
        req = requests.get(url, params=parameters, timeout=20)
        req.raise_for_status()
        return req.json()

    def get_all_albums(self) -> List[str]:
        albums = self.call_sonic("getAlbumList", type="alphabeticalByName", size=500)[
            "subsonic-response"
        ]["albumList"]["album"]
        return [a["album"] for a in albums]

    def get_songs_for_album(self, album_name: str) -> Tuple[str, List[Song]]:
        output = self.call_sonic("getAlbumList", type="alphabeticalByName", size=500)[
            "subsonic-response"
        ]
        albums = output["albumList"]["album"]
        songs = []
        closest = difflib.get_close_matches(
            album_name,
            # This truncation is a hack, but get_close_matches doesn't handle very
            # dissimilar string length well. We essentially assume that the user
            # was lazy and just typed the beginning of the album name, which works
            # actually really well. It is a good enough heuristic for now.
            [a["album"][: len(album_name) + 3] for a in albums],
            1,
        )
        if not closest:
            raise AlbumNotFoundException(album_name)

        for album in albums:
            if album["album"][: len(album_name) + 3] == closest[0]:
                cover_url, cover_params = self.make_sonic_url(
                    "getCoverArt", id=album["coverArt"]
                )
                theid = album["id"]
                data = self.call_sonic("getAlbum", id=theid)["subsonic-response"][
                    "album"
                ]["song"]
                for s in data:
                    strurl, params = self.make_sonic_url("stream", id=s["id"])
                    songs.append(
                        Song(
                            s["title"],
                            s["album"],
                            s["artist"],
                            strurl + "?" + urlencode(params),
                            s["contentType"],
                            cover_url + "?" + urlencode(cover_params),
                        )
                    )
                return album["album"], songs

        raise AlbumNotFoundException(album_name)



from contextlib import contextmanager, redirect_stdout
from dataclasses import dataclass
from enum import Enum
from io import BytesIO
from queue import Empty, Queue
from threading import Thread
from typing import Any, BinaryIO, Generator, List

import requests as r

# Feeling bad about it, but pygame always display a welcome
# message which is completely out of place on a CLI music player.
with redirect_stdout(None):
    from pygame import event, NOEVENT
    from pygame.locals import USEREVENT
    from pygame.display import init as display_init
    from pygame.mixer import music
    from pygame.mixer import init as mixer_init

from castme.config import Config
from castme.messages import debug as msg_debug
from castme.player import Backend, NoSongsToPlayException
from castme.song import Song

STOP_EVENT = USEREVENT + 1


def get_song(song: Song) -> BinaryIO:
    response = r.get(song.url, timeout=10)
    response.raise_for_status()
    return BytesIO(response.content)


def debug(msg):
    msg_debug("local", msg)


@dataclass
class Message:
    class Type(Enum):
        VOLUME_SET = 1
        VOLUME_DELTA = 2
        PLAY_PAUSE = 3
        FORCE_PLAY = 4
        STOP = 5
        EXIT = 6
        PLAY = 7

    @staticmethod
    def playpause():
        return Message(Message.Type.PLAY_PAUSE, None)

    @staticmethod
    def stop():
        return Message(Message.Type.STOP, None)

    @staticmethod
    def exit():
        return Message(Message.Type.EXIT, None)

    @staticmethod
    def force_play():
        return Message(Message.Type.FORCE_PLAY, None)

    type: Type
    # This is ugly but it will do for now. Poor man's tagged union
    payload: Any


class State(Enum):
    PLAYING = 1
    PAUSED = 2
    STOPPED = 3


def play_next(songs: List[Song]) -> bool:
    if songs:
        debug(f"Playing {songs[0].title}")
        music.load(get_song(songs[0]))
        music.play()
        return True
    return False


def pygame_loop(queue: Queue[Message], songs: List[Song]):  # noqa: PLR0912
    """Pygame is not thread-safe. All the api calls needs to be done on the
    same thread, expecially the event management code."""
    mixer_init()
    display_init()
    music.set_endevent(STOP_EVENT)

    state = State.STOPPED

    while True:
        try:
            message = queue.get(timeout=0.1)
            debug(f"loop - Received message {message}")
            match message.type:
                case Message.Type.VOLUME_SET:
                    music.set_volume(message.payload)
                case Message.Type.VOLUME_DELTA:
                    music.set_volume(music.get_volume() + message.payload)
                case Message.Type.STOP:
                    state = State.STOPPED
                    music.stop()
                case Message.Type.PLAY_PAUSE:
                    if state == State.STOPPED:
                        if play_next(songs):
                            state = State.PLAYING
                    elif state == State.PAUSED:
                        music.unpause()
                        state = State.PLAYING
                    elif state == State.PLAYING:
                        music.pause()
                        state = State.PAUSED
                case Message.Type.FORCE_PLAY:
                    if play_next(songs):
                        state = State.PLAYING
                case Message.Type.EXIT:
                    return
        except Empty:
            pass

        if (pygame_event := event.poll()).type != NOEVENT:
            debug(f"Event: {pygame_event}")
            if pygame_event.type == STOP_EVENT:
                if state == State.PLAYING:
                    # The channel have stopped _and_ we are now playing the queued song. It is time
                    # to move on to the next song
                    if songs:
                        songs.pop(0)

                    if songs and play_next(songs):
                        debug("Channel was not busy, played the next song")
                        state = State.PLAYING
                    else:
                        debug("Channel was not busy, nothing to play")
                        state = State.STOPPED


class LocalBackendImpl(Backend):
    def __init__(self, songs: List[Song]):
        self.songs = songs
        self.queue: Queue[Message] = Queue()
        self.pygame_thread = Thread(
            target=pygame_loop,
            args=(
                self.queue,
                self.songs,
            ),
        )
        self.pygame_thread.start()

    def close(self):
        self.queue.put(Message.exit())
        self.pygame_thread.join()

    def playpause(self):
        self.queue.put(Message.playpause())

    def force_play(self):
        if not self.songs:
            raise NoSongsToPlayException()
        self.queue.put(Message.force_play())

    def volume_set(self, value: float):
        self.queue.put(Message(Message.Type.VOLUME_SET, value))

    def volume_delta(self, value: float):
        self.queue.put(Message(Message.Type.VOLUME_DELTA, value))

    def stop(self):
        self.queue.put(Message.stop())


@contextmanager
def backend(_config: Config, songs: List[Song]) -> Generator[Backend, None, None]:
    local = LocalBackendImpl(songs)
    try:
        yield local
    finally:
        local.close()



from contextlib import contextmanager
from typing import Generator, List

from pychromecast import Chromecast, get_listed_chromecasts  # type: ignore
from pychromecast.controllers.media import (  # type: ignore
    MediaController,
    MediaStatus,
    MediaStatusListener,
)

from castme.config import Config
from castme.messages import debug as msg_debug
from castme.messages import error
from castme.player import Backend, NoSongsToPlayException
from castme.song import Song


def debug(msg: str):
    msg_debug("chromecast", msg)


class ChromecastBackend(Backend):
    def __init__(self, config: Config, songs: List[Song]):
        self.chromecast_friendly_name = config.chromecast_friendly_name
        self.songs = songs
        self.chromecast = find_chromecast(self.chromecast_friendly_name)
        self.mediacontroller = self.chromecast.media_controller
        self.chromecast.wait()
        self.mediacontroller.register_status_listener(
            MyChromecastListener(songs, self.mediacontroller)
        )

    def force_play(self):
        debug("Force play")
        if self.songs:
            play_on_chromecast(self.songs[0], self.mediacontroller)
        else:
            raise NoSongsToPlayException()

    def playpause(self):
        debug("playpause")
        if self.mediacontroller.status.player_is_paused:
            self.mediacontroller.play()
        elif self.mediacontroller.status.player_is_idle:
            self.force_play()
        elif self.mediacontroller.status.player_is_playing:
            self.mediacontroller.pause()

    def volume_set(self, value: float):
        debug(f"volume set {value}")
        self.chromecast.set_volume(value)

    def volume_delta(self, value: float):
        debug(f"volume delta {value}")
        if value > 0:
            self.chromecast.volume_up(value)
        else:
            self.chromecast.volume_down(-value)

    def stop(self):
        debug("stop")
        if self.mediacontroller.is_active:
            self.mediacontroller.stop()

    def close(self):
        debug("close")
        self.stop()


class ChromecastNotFoundException(Exception):
    def __init__(self, keyword: str):
        self.keyword = keyword

    def __str__(self):
        return f"Chromecast named {self.keyword} not found"


class MyChromecastListener(MediaStatusListener):
    def __init__(self, songs: List[Song], media_controller: MediaController):
        self.songs = songs
        self.media_controller = media_controller

    def new_media_status(self, status: MediaStatus):
        if status.player_is_idle and status.idle_reason == "FINISHED":
            if self.songs:
                self.songs.pop(0)
            if self.songs:
                play_on_chromecast(self.songs[0], self.media_controller)

    def load_media_failed(self, item: int, error_code: int):
        """Called when load media failed."""
        error(f"Error loading media, error code: {error_code}")


def find_chromecast(label: str) -> Chromecast:
    chromecasts, _ = get_listed_chromecasts(friendly_names=[label])
    if not chromecasts:
        raise ChromecastNotFoundException(label)

    return chromecasts[0]


def play_on_chromecast(song: Song, controller: MediaController):
    metadata = dict(
        # 3 is the magic number for MusicTrackMediaMetadata
        # see https://developers.google.com/cast/docs/media/messages
        metadataType=3,
        albumName=song.album_name,
        title=song.title,
        artist=song.artist,
    )
    debug(f"Playing {song.title} @ {song.url}")
    controller.play_media(
        song.url,
        content_type=song.content_type,
        title=song.title,
        media_info=metadata,
        thumb=song.album_art,
    )


@contextmanager
def backend(config: Config, songs: List[Song]) -> Generator[Backend, None, None]:
    chromecast = ChromecastBackend(config, songs)
    try:
        yield chromecast
    finally:
        chromecast.close()



import argparse
import cmd
import os
import shutil
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path
from shutil import get_terminal_size
from sys import exit as sys_exit
from typing import Dict, List

from castme.backends.chromecast import backend as chromecast_backend
from castme.backends.local import backend as local_backend
from castme.config import Config
from castme.messages import debug_mode_enabled, enable_debug_mode, error, message
from castme.player import Backend, NoSongsToPlayException
from castme.song import Song
from castme.subsonic import AlbumNotFoundException, SubSonic


class InvalidBackend(Exception):
    def __init__(self, invalid_backend_name: str):
        self.invalid_name = invalid_backend_name

    def __str__(self):
        return f"Invalid backend name {self.invalid_name}"


def castme_version():
    try:
        return version("castme")
    except PackageNotFoundError:
        return "unknown"


SUBSONIC_APP_ID = "castme"


class CastMeCli(cmd.Cmd):
    def __init__(
        self,
        subsonic: SubSonic,
        targets: Dict[str, Backend],
        default_backend: str,
        songs: List[Song],
    ):
        super().__init__()
        self.min_column_width = 50
        self.subsonic = subsonic
        self.songs = songs
        self.targets = targets
        if default_backend not in targets:
            raise InvalidBackend(default_backend)

        self.current_target = targets[default_backend]
        message(f"Currently playing on {default_backend}")
        self.update_prompt(default_backend)

    def update_prompt(self, label: str):
        self.prompt = f"[{label}] >> "

    def do_list(self, _line: str):
        """List all the albums available (alias: l)"""
        term_cols, term_rows = get_terminal_size()
        albums = self.subsonic.get_all_albums()
        number_of_columns = term_cols // self.min_column_width or 1
        # We can get some extra chars by dispatching the remainder characters to
        # each column
        column_width = (
            self.min_column_width
            + (term_cols % self.min_column_width) // number_of_columns
        )
        text_width_fmt = str(column_width - 2)  # 2 chars of padding
        # We want to truncate the string to the text width
        format_string_album = "{:" + text_width_fmt + "." + text_width_fmt + "}"

        while albums:
            lines_printed = 1
            # We print line by line
            while albums and lines_printed < term_rows:
                displayed_albums = albums[:number_of_columns]
                albums = albums[number_of_columns:]
                # This line concatenate N format string and then format the result with N album names
                message(
                    "".join(([format_string_album] * len(displayed_albums))).format(
                        *displayed_albums
                    )
                )
                lines_printed += 1

            if albums:
                input(" .... Press <Enter> to continue ....")

    def emptyline(self):
        pass

    def do_switch(self, line):
        """Switch to another backend. Without argument list the available
        backends. (alias: s)"""
        if not line:
            message(f"Available targets: {', '.join(self.targets.keys())}")
            return

        if line in self.targets:
            self.current_target.stop()
            self.current_target = self.targets[line]
            if self.songs:
                self.current_target.force_play()

            self.update_prompt(line)
        else:
            error(f"Could not find target {line}")

    def do_clear(self, _line: str):
        """Clear the queue and stop the music (alias: c)"""
        self.songs.clear()
        self.current_target.stop()

    def do_queue(self, line: str):
        """Queue an album. The argument to that command will be matched against all
        albums on the device and the best matching one will be played (alias: q).
        """
        if not line:
            for idx, s in enumerate(self.songs):
                message(f"{1 + idx:2} {s}")
            return
        try:
            start_empty = len(self.songs) == 0
            name, songs = self.subsonic.get_songs_for_album(line)
            message(f"Queueing {name}")
            self.songs.extend(songs)
            if start_empty:
                self.current_target.force_play()
        except AlbumNotFoundException as e:
            error(str(e))

    def do_playpause(self, _line: str):
        """play/pause the song (alias: pp)"""
        self.current_target.playpause()

    def do_next(self, _line: str):
        """Skip to the next song (alias: n)"""
        if self.songs:
            self.songs.pop(0)
        try:
            self.current_target.force_play()
        except NoSongsToPlayException:
            error("No songs in the queue")

    def do_volume(self, line: str):
        """Set or change the volume. Valid values are between 0 and 100 (alias: v)
        +VALUE: Increase the volume by VALUE
        -VALUE: Decrease the volume by VALUE
        VALUE: Set the volume to VALUE
        """
        try:
            value = float(line) / 100.0
        except ValueError:
            error("Error converting the value into a number")
            return

        if line.startswith("+") or line.startswith("-"):
            self.current_target.volume_delta(value)
        else:
            self.current_target.volume_set(value)

    def do_quit(self, _line: str):
        """Exit the application (alias: x or Ctrl-D)"""
        self.current_target.stop()
        return True

    def precmd(self, line: str) -> str:
        potential_alias = line.split(" ")[0]
        aliases = {
            "pp": "playpause",
            "l": "list",
            "n": "next",
            "q": "queue",
            "v": "volume",
            "c": "clear",
            "x": "quit",
            "s": "switch",
            "EOF": "quit",  # Set by Cmd itself on Ctrl-C
        }
        if potential_alias in aliases:
            return line.replace(potential_alias, aliases[potential_alias], 1)
        else:
            return line


def main():
    parser = argparse.ArgumentParser("CastMe")
    parser.add_argument("--config", help="Set the configuration file to use")
    parser.add_argument(
        "--init",
        help="create an empty configuration file in ~/.config . You can override its location with --config",
        action="store_true",
    )
    parser.add_argument("--version", action="store_true", help="Print version and exit")
    parser.add_argument("--debug", action="store_true", help="print debugging messages")
    parser.add_argument("backend", nargs="?")
    args = parser.parse_args()
    config_path = args.config

    if args.debug:
        enable_debug_mode()

    if args.version:
        message(f"Version: {castme_version()}")
        return

    try:
        if args.init:
            config_path = os.path.expanduser(args.config or "~/.config/castme.toml")
            if os.path.exists(config_path):
                error(
                    f"The configuration file {config_path} already exist, bailing out..."
                )
                sys_exit(1)
            shutil.copy(
                Path(os.path.dirname(__file__), "assets/castme.toml.template"),
                config_path,
            )
            message(
                f"Configuration initialized in {config_path}, please edit it before starting castme again"
            )
            sys_exit(0)

        config = Config.load(config_path)
        subsonic = SubSonic(
            SUBSONIC_APP_ID, config.user, config.password, config.subsonic_server
        )

        songs_queue: List[Song] = []

        with (
            chromecast_backend(config, songs_queue) as chromecast,
            local_backend(config, songs_queue) as local,
        ):

            cli = CastMeCli(
                subsonic,
                {"chromecast": chromecast, "local": local},
                args.backend or config.default_backend,
                songs_queue,
            )
            cli.cmdloop()
    except Exception as e:
        if debug_mode_enabled():
            raise
        else:
            error(str(e))
            sys_exit(1)


if __name__ == "__main__":
    main()



import os.path
import tomllib
from dataclasses import dataclass
from pathlib import PurePath
from typing import Optional

from castme.messages import debug


class ConfigNotFoundException(Exception):
    pass


@dataclass
class Config:
    user: str
    password: str
    subsonic_server: str
    chromecast_friendly_name: str
    default_backend: str

    @classmethod
    def load(cls, file_path: Optional[PurePath | str] = None) -> "Config":
        def _load(path: PurePath | str) -> "Config":
            debug("config", f"Loading config from {path}")
            with open(path, "rb") as fd:
                data = tomllib.load(fd)
            return cls(**data)

        if file_path:
            return _load(file_path)

        for f in [
            "castme.toml",
            os.path.expanduser("~/.config/castme.toml"),
            "/etc/castme.toml",
        ]:
            if os.path.isfile(f):
                return _load(f)

        raise ConfigNotFoundException()
